/*-------------------------------------------------------------------------*/
/*  J3W ver 6.50  3D Animation Kit                                         */
/*  spaceh3d.cpp  01/02/2006                                               */
/*  Copyright (C) 1996 - 2006  Jun Mizutani  <mizutani.jun@nifty.ne.jp>    */
/*                      All rights reserved.                               */
/*                                                                         */
/*   This file is part of the J3W 3D Animation Kit, and is covered under   */
/*  the terms of the GNU General Public License, version 2. This file has  */
/*  NO WARRANTY. See file COPYING for copyright details.                   */
/*                                                                         */
/*-------------------------------------------------------------------------*/

#include <iostream>
#include <stdlib.h>
#include <math.h>
#include "spaceh3d.h"

using namespace std;

ClipType   FrontClip, BackClip, LeftClip, RightClip, TopClip, BottomClip;

TSpaceH3D::TSpaceH3D(int objNo, int front, int back):
                    light_dir(100, 0, 0)
{
    maxObj = objNo;
    pEye = 0;
    pLight = 0;
    ObjArray = new THObj3D*[maxObj];
    for(int i=0; i<objNo; i++) ObjArray[i] = 0;
    ObjCount = 0;
    VertCount = 0;
    angle = 0;
    parallel = 1;
    ColorNum = ColNum;
    Shade    = ColWidth;
    ClippedVertex = new Vector[MAX2DVERTEX];
    pglist = new PolygonType[MAXPOLYGON];
    pglistmax = 0;
    Polygon2D = new Vector2D[MAX2DVERTEX];
    max2D = 0;
    set_clipplane(front, back);
}

TSpaceH3D::~TSpaceH3D()
{
    reset_space();
    delete ObjArray;
    delete ClippedVertex;
    delete pglist;
    delete Polygon2D;
}

void TSpaceH3D::set_angle(int a)
{
   if (a == 0) angle = 0;
   else angle = 1;
}

void TSpaceH3D::reset_space()
{
    for(int i=0; i<maxObj; i++) delete_object(i);
}

void TSpaceH3D::delete_object(int n)
{
    if ((ObjCount > 0) && (n < maxObj) && (ObjArray[n] != 0)) {
        delete ObjArray[n];
        ObjArray[n] = 0;
        ObjCount--;
  }
}

int TSpaceH3D::append_object(int vn, int pn)
{
    if (ObjCount < maxObj) {
        int i = 0;
        while ((i<(maxObj-1)) && (ObjArray[i] != 0)) i++;
        ObjArray[i] = new THObj3D(vn, pn, 0);
        ObjCount++;
        return i;
    } else return -1;
}

int TSpaceH3D::append_child_object(int n, int vn, int pn)
{
    if (ObjCount < maxObj) {
        int i = 0;

        while ((i<(maxObj-1)) && (ObjArray[i] != 0)) i++;

        if (ObjArray[n]->child == 0) {
            ObjArray[i] = new THObj3D(vn, pn, ObjArray[n]);
            ObjArray[n]->child = ObjArray[i];
        } else {
            THObj3D *p = ObjArray[n]->child;
            while (p->brother != 0) p = p->brother;

            ObjArray[i] = new THObj3D(vn, pn, ObjArray[n]);
            p->brother = ObjArray[i];
        }
        ObjCount++;
        return i;
    } else return -1;
}

int TSpaceH3D::count_vertex()
{
    int n = 0;
    for(int i=0; i<maxObj; i++)
         if (ObjArray[i]) n += ObjArray[i]->Polygon.Vertex.GetCount();
    return n;
}

int TSpaceH3D::count_polygon()
{
    int n = 0;
    for(int i=0; i<maxObj; i++)
        if (ObjArray[i]) n += ObjArray[i]->Polygon.GetCount();
    return n;
}

void TSpaceH3D::display(int delta)
{
    if (pEye!=0) {

        for(int i = 0; i<maxObj; i++)
            if ((ObjArray[i] != 0) && (ObjArray[i]->parent == 0))
                ObjArray[i]->SetNodeWorldMatrix(delta);

        for(int i=0; i<maxObj; i++)
            if (ObjArray[i] != 0) ObjArray[i]->show(pEye);
        if (0 < maxObj) {
            Clip3D_PolygonList();
            Polygon3Dto2D();
            ShadePolygon();
            if (pglistmax > 0) DepthSort();
        }
        if (PScreen->GetColorMode() == TRUECOLOR) {
            for(int i=0; i<max2D; i++)
                if (pglist[i].color >= 0) {
                    if (pglist[i].n == 2)
                        PScreen->line(Polygon2D[ pglist[i].index ],
                            Polygon2D[ pglist[i].index+1 ], pglist[i].color);
                    else {
                        int col = pglist[i].color;
                        if ((col == 255) || ((col > 15) && (col < 43))) {
                            PScreen->polygon(&Polygon2D[pglist[i].index],
                                              pglist[i].n, pglist[i].r+SYSCOLOR);
                        } else {
                            PScreen->polygon_rgb(&Polygon2D[pglist[i].index],
                                                  pglist[i].n, pglist[i].r,
                                                  pglist[i].g, pglist[i].b);
                        }
                    }
                }
        } else {
            for(int i=0; i<max2D; i++)
                if (pglist[i].color >= 0) {
                    if (pglist[i].n == 2)
                        PScreen->line(Polygon2D[ pglist[i].index ],
                            Polygon2D[ pglist[i].index+1 ], pglist[i].color);
                    else {
                        int col = pglist[i].r;
                        if ((col == 255) || ((col > 15) && (col < 43))) {
                            PScreen->polygon(&Polygon2D[pglist[i].index],
                                              pglist[i].n, pglist[i].r);
                        } else {
                            PScreen->polygon(&Polygon2D[pglist[i].index],
                                     pglist[i].n, pglist[i].r);
                        }
                    }
                }
        }
    }
}

void TSpaceH3D::set_clipplane(int front, int back)
{
    FrontClip.kx = 1;   FrontClip.ky = 0;
    FrontClip.kz = 0;   FrontClip.m = -front;
    BackClip.kx = -1;   BackClip.ky = 0;
    BackClip.kz = 0;    BackClip.m = back;
    LeftClip.kx = 1;    LeftClip.ky = 1;
    LeftClip.kz = 0;    LeftClip.m = 0;
    RightClip.kx = 1;   RightClip.ky = -1;
    RightClip.kz = 0;   RightClip.m = 0;
    TopClip.kx = 1;     TopClip.ky = 0;
    TopClip.kz = 1;     TopClip.m  = 0;
    BottomClip.kx = 1;  BottomClip.ky = 0;
    BottomClip.kz = -1; BottomClip.m  = 0;
}

void  TSpaceH3D::ClipLine(Vector &p1, Vector &p2, ClipType &clip, Vector &p)
{
    int dx = p2.x - p1.x;
    int dy = p2.y - p1.y;
    int dz = p2.z - p1.z;
    double numerator = clip.kx*p1.x + clip.ky*p1.y + clip.kz*p1.z + clip.m;
    double denominator = -clip.kx*dx - clip.ky*dy - clip.kz*dz;
    p.x = iround(dx * numerator / denominator) + p1.x;
    p.y = iround(dy * numerator / denominator) + p1.y;
    p.z = iround(dz * numerator / denominator) + p1.z;
}

inline int TSpaceH3D::visible(Vector &p, ClipType &clip)
{
    return (clip.kx*p.x + clip.ky*p.y + clip.kz*p.z + clip.m > 0);
}

int TSpaceH3D::Clip_polygon(int n, Vector *polygon_in, Vector* polygon_out,
                        ClipType &clipEq)
{
    Vector p0, p1, p2 ,p;
    int p0vis, p1vis, p2vis;
    int OutCount = 0;

    if (n == 2) {
        p1 = polygon_in[0];
        p2 = polygon_in[1];
        if (visible(p1 , clipEq)) {
            polygon_out[OutCount++] = p1;
            if (!visible(p2, clipEq)) {
                ClipLine(p1,p2,clipEq,p);
                polygon_out[OutCount++] = p;
            } else polygon_out[OutCount++] = p2;
        } else {
            if (visible(p2, clipEq)) {
                ClipLine(p1,p2,clipEq,p);
                polygon_out[OutCount++] = p;
                polygon_out[OutCount++] = p2;
            }
        }
        return OutCount;
    }

    p0 = polygon_in[0];
    p0vis = visible(p0 , clipEq);
    p1 = p0;
    p1vis = p0vis;
    for ( int i=1; i<n; i++) {
        p2 = polygon_in[i];
        p2vis = visible(p2 , clipEq);
        if (p1vis && p2vis)
            polygon_out[OutCount++] = p1;
        else if (p1vis) {
            polygon_out[OutCount++] = p1;
            ClipLine(p1,p2,clipEq,p);
            polygon_out[OutCount++] = p;
        } else if (p2vis) {
            ClipLine(p1,p2,clipEq,p);
            polygon_out[OutCount++] = p;
        }
        p1 = p2;
        p1vis = p2vis;
    }
    if (p1vis && p0vis)
        polygon_out[OutCount++] = p1;
    else if (p1vis) {
        polygon_out[OutCount++] = p1;
        ClipLine(p1,p0,clipEq,p);
        polygon_out[OutCount++] = p;
    } else if (p0vis) {
        ClipLine(p1,p0,clipEq,p);
        polygon_out[OutCount++] = p;
    }
    return OutCount;
}

void TSpaceH3D::Clip_volume(int &n, Vector *polygon1)
{
    Vector polygon2[POLYGONVERTEX];
    n = Clip_polygon(n,polygon1, polygon2, BackClip);
    if (n > 0) n = Clip_polygon(n, polygon2, polygon1, FrontClip);
    if (n > 0) n = Clip_polygon(n, polygon1, polygon2, LeftClip);
    if (n > 0) n = Clip_polygon(n, polygon2, polygon1, RightClip);
    if (n > 0) n = Clip_polygon(n, polygon1, polygon2, TopClip);
    if (n > 0) n = Clip_polygon(n, polygon2, polygon1, BottomClip);
}

void TSpaceH3D::Clip3D_PolygonList()
{
    int n;
    Vector va[POLYGONVERTEX];
    Vector ve;
    Polygon3D *p;
    TPolygons *po;
    double x0,y0,z0,x1,y1,z1,x2,y2,z2;

    double wh_ratio = double(PScreen->get_width()) / double(PScreen->get_height());
    pglistmax = 0;
    VertCount = 0;
    for(int i=0; i<maxObj; i++)
        if (ObjArray[i] != 0) {
            po = &ObjArray[i]->Polygon;
            for(int j = 0; j<po->GetCount(); j++)
                if (po->pg[j] != 0) {
                    n = po->pg[j]->n;
                    p = po->pg[j];
                    int kk = 0;

                    if (p->index[0] < 0) {
                        THObj3D* pp = ObjArray[i]->parent;
                        if (pp != 0) {
                            int m = -(p->index[0]);
                            if (m < pp->Polygon.Vertex.GetCount())
                                ve = pp->Polygon.Vertex.vert[m]->eye;
                        }
                    } else {
                        ve = po->Vertex.vert[p->index[0]]->eye;
                    }

                    for(int k=0; k<n; k++) {
                        if (p->index[k] < 0) {
                            THObj3D* pp = ObjArray[i]->parent;
                            if (pp != 0) {
                                int m = -(p->index[k]);
                                if (m < pp->Polygon.Vertex.GetCount()) {
                                    va[kk] = pp->Polygon.Vertex.vert[m]->eye;
                                    va[kk].z = iround(va[kk].z * wh_ratio);
                                    if (angle == 1) va[kk].x = va[kk].x >> 1;
                                    kk++;
                                }
                            }
                        } else {
                            va[kk] = po->Vertex.vert[p->index[k]]->eye;
                            va[kk].z = iround(va[kk].z * wh_ratio);
                            if (angle == 1) va[kk].x = va[kk].x >> 1;
                            kk++;
                        }
                    }

                    n = kk;

                    /* when parent's vertex is referred, calc normal vector */
                    Vector v[3];

                    if ((n > 2) && (p->f_normal == 0)) {
                        for(int h=0; h<3; h++) {        /* for every vertex */
                            if (p->index[h] < 0) {
                                THObj3D* pp = ObjArray[i]->parent;
                                if (pp != 0) {
                                    v[h] = pp->Polygon.Vertex.vert[-(p->index[h])]->eye;
                                }
                            } else {
                                v[h] = po->Vertex.vert[p->index[h]]->eye;
                            }
                        }
                        x0 = v[0].x; y0 = v[0].y; z0 = v[0].z;
                        x1 = v[1].x; y1 = v[1].y; z1 = v[1].z;
                        x2 = v[2].x; y2 = v[2].y; z2 = v[2].z;
                        p->enormal.x = (y1-y0)*(z2-z1) - (z1-z0)*(y2-y1);
                        p->enormal.y = (z1-z0)*(x2-x1) - (x1-x0)*(z2-z1);
                        p->enormal.z = (x1-x0)*(y2-y1) - (y1-y0)*(x2-x1);
                        p->enormal.normalize();
                    }

                    Vector_d vd(ve);
                    vd.normalize();
                    vd.neg();
                    if (vd.dot_product(p->enormal) > 0.0) {
                        Clip_volume(n, va);
                        if (n != 0) {
                            pglist[pglistmax].index = VertCount;
                            pglist[pglistmax].n = n;
                            pglist[pglistmax].color = p->color;
                            pglist[pglistmax].reflection = p->reflection;
                            pglist[pglistmax].ambient = p->ambient;
                            pglist[pglistmax].normal = p->enormal;
                            pglist[pglistmax].off_center = p->off_center;
                            for(int k=0; k<n; k++) ClippedVertex[VertCount+k] = va[k];
                            if (pglistmax < (MAXPOLYGON-1)) pglistmax++;
                            VertCount += n;
                        }
                    }
            }
        }
}

int SortCompare(const void *p, const void *q )
{
    if ( ((PolygonType *)p)->off_center == ((PolygonType *)q)->off_center)
        return 0;
    if ( ((PolygonType *)p)->off_center > ((PolygonType *)q)->off_center)
        return -1;
    else return 1;
}

void TSpaceH3D::DepthSort()
{
    qsort(pglist, pglistmax, sizeof(PolygonType), SortCompare);
}

void TSpaceH3D::Polygon3Dto2D()
{
    int  k, m;
    double xx, yy, zz;

    max2D = pglistmax;
    for (int i = 0; i<pglistmax; i++) {
        k = pglist[i].n;
        m = pglist[i].index;
        for(int j=0; j<k; j++) {
            xx = ClippedVertex[m+j].x;
            yy = ClippedVertex[m+j].y;
            zz = ClippedVertex[m+j].z;
            Polygon2D[m+j].x = int((yy / xx + 1.0)*(PScreen->get_width()-2) /2)
                                        + PScreen->get_left()+1;
            Polygon2D[m+j].y = int((1.0 + zz / xx)*(PScreen->get_height()-2) /2)
                                        + PScreen->get_top()+1;
        }
    }
}

double Power(double X, int N)
{
  if (X == 0) return 0;
  return exp(N * log(X));
}

void shading(double cos_theta, double specl, PolygonType *p)
{
    int specular = int(0.5 + specl * HIGHLIGHT * p->reflection * 255);
    if (cos_theta < 0) cos_theta = 0;
    cos_theta = cos_theta * (1.0 - p->ambient) + p->ambient;

    double Y  =  0.2990 * p->r + 0.5870 * p->g + 0.1140 * p->b;
    double Cb = -0.1687 * p->r - 0.3313 * p->g + 0.5000 * p->b;
    double Cr =  0.5000 * p->r - 0.4187 * p->g - 0.0813 * p->b;

    Y *= cos_theta;

    p->r = int(Y + 1.4020*Cr);
    p->g = int(Y - 0.3441*Cb - 0.7141*Cr);
    p->b = int(Y + 1.7720*Cb);

    if (p->r < 0) p->r = 0;
    if (p->g < 0) p->g = 0;
    if (p->b < 0) p->b = 0;

    p->r += specular;
    p->g += specular;
    p->b += specular;

    if (p->r > 255) p->r = 255;
    if (p->g > 255) p->g = 255;
    if (p->b > 255) p->b = 255;
}

void TSpaceH3D::ShadePolygon()
{
    int  k, m;
    double cos_theta;

    double hw_ratio = double(PScreen->get_height()) / double(PScreen->get_width());

    max2D = pglistmax;
    for (int i = 0; i<pglistmax; i++) {
        k = pglist[i].n;
        m = pglist[i].index;

        if (k == 2) {
            if (pglist[i].color >= 0x100) {

                pglist[i].color &= 0xFF;
            } else if ((pglist[i].color > 15) && (ColorNum <= 16)) {

                if (pglist[i].color < 43)
                    pglist[i].color = ColNum*ColWidth + pglist[i].color -16 + SYSCOLOR;
            } else  pglist[i].color = (pglist[i].color) * Shade + SYSCOLOR;
        } else {

            int sum_x = 0;
            int sum_y = 0;
            int sum_z = 0;
            for(int j=0; j < k; j++) {
              sum_x += ClippedVertex[m + j].x;
              sum_y += ClippedVertex[m + j].y;
              sum_z += int(ClippedVertex[m + j].z * hw_ratio + 0.5);
            }
            sum_x /= k;
            sum_y /= k;
            sum_z /= k;
            pglist[i].center_g = Vector(sum_x, sum_y, sum_z);

            if (angle == 1) pglist[i].center_g.x *= 2;

            pglist[i].off_center += pglist[i].center_g.x;

            Vector_d center = Vector(0,0,0) - pglist[i].center_g;
            center.normalize();
            cos_theta = center.dot_product(pglist[i].normal);

            if (cos_theta < 0) pglist[i].color = -1;
            else if (pglist[i].color >= 256) {

                pglist[i].color &= 0xFF;
            } else if ((pglist[i].color > 15) && (ColorNum <= 16)) {

                if (pglist[i].color < 43)
                    pglist[i].r = ColNum*ColWidth + pglist[i].color -16;
            } else {
                if (PScreen->GetColorMode() == PSEUDOCOLOR) {

                    pglist[i].r = (pglist[i].color) * Shade
                            + int((1-cos_theta)*(Shade-1)+0.5) + SYSCOLOR;
                } else {

                    int pal_color = pglist[i].color * Shade;
                    pglist[i].r = MyPalette[ pal_color].r;
                    pglist[i].g = MyPalette[ pal_color].g;
                    pglist[i].b = MyPalette[ pal_color].b;

                    Vector_d light_vec;
                    if (parallel) {
                        if (pLight == 0) {
                            light_vec = light_dir;
                            light_vec.neg();
                        } else {

                            Vector dir = pLight->GetDirection_world(light_dir);
                            light_vec = pEye->GetDirection_eye(dir);
                            light_vec.neg();
                        }
                     } else {
                        if (pLight == 0) {

                            light_vec = center;
                        } else {
                            Vector light;
                            int h, p, b;
                            pLight->GetWorldPosition(light, h, p, b);
                            light = pEye->WorldToLocal(light);
                            light_vec = light - pglist[i].center_g;
                        }
                    }
                    light_vec.normalize();

                    double cos_ln = light_vec.dot_product(pglist[i].normal);
                    Vector_d reflection = 2 * cos_ln * pglist[i].normal - light_vec;

                    double cos_nr = center.dot_product(reflection);
                    if (cos_nr < 0) cos_nr = 0;
                    double specular = Power(cos_nr, POWER);
                    shading(cos_ln, specular, &pglist[i]);
                }
            }
        }
    }
}

